"""
This folder contains one file per each supported data file.

Each file contains the binary load and preprocessing functions needed for
the specific format. Functions defined here are used by loader functions
in `loaders.py` to properly initialize a `Data()` variable.
"""